#include <algorithm>
#include <cmath>
#include <msp/fs/path.h>
#include <msp/gl/framebuffer.h>
#include <msp/gl/immediate.h>
#include <msp/gl/matrix.h>
#include <msp/gl/renderer.h>
#include <msp/gl/texture2d.h>
#include <msp/io/print.h>
#include "overlay.h"
#include "track.h"
#include "tracktype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Overlay3D::Overlay3D(const GL::Font &f):
	font(f)
{ }

Overlay3D::~Overlay3D()
{
	for(map<string, GL::Mesh *>::iterator i=graphics.begin(); i!=graphics.end(); ++i)
		delete i->second;
	for(map<const Object3D *, Icon *>::iterator i=icons.begin(); i!=icons.end(); ++i)
		delete i->second;
}

void Overlay3D::set_label(const Object3D &track, const string &label)
{
	get_icon(track).label = label;
	update_icon(get_icon(track));
}

void Overlay3D::add_graphic(const Object3D &track, const string &grf_name)
{
	const GL::Mesh *grf = get_graphic(grf_name);
	if(!grf)
		return;

	Icon &icon = get_icon(track);
	if(find(icon.graphics.begin(), icon.graphics.end(), grf)==icon.graphics.end())
		icon.graphics.push_back(grf);

	update_icon(icon);
}

void Overlay3D::remove_graphic(const Object3D &track, const string &grf_name)
{
	const GL::Mesh *grf = get_graphic(grf_name);
	Icon &icon = get_icon(track);
	icon.graphics.erase(remove(icon.graphics.begin(), icon.graphics.end(), grf), icon.graphics.end());

	update_icon(icon);
}

void Overlay3D::clear_graphics(const Object3D &track)
{
	get_icon(track).graphics.clear();
	update_icon(get_icon(track));
}

void Overlay3D::clear(const Object3D &track)
{
	map<const Object3D *, Icon *>::iterator i = icons.find(&track);
	if(i!=icons.end())
	{
		delete i->second;
		icons.erase(i);
	}
}

void Overlay3D::render(GL::Renderer &renderer, const GL::Tag &) const
{
	unsigned fb_width = GL::Framebuffer::current()->get_width();
	unsigned fb_height = GL::Framebuffer::current()->get_height();

	GL::Renderer::Push push(renderer);
	GL::MatrixStack::Push push_proj(GL::MatrixStack::projection());

	GL::MatrixStack::projection() = GL::Matrix::scaling(2.0/fb_width, 2.0/fb_height, 1.0);

	glLineWidth(1);

	int size = int(font.get_native_size()+0.5);
	float spacing = round(size*1.1)/size;
	float baseline = round((0.5-font.get_ascent()*0.5-font.get_descent()*0.5)*size)/size;

	for(map<const Object3D *, Icon *>::const_iterator i=icons.begin(); i!=icons.end(); ++i)
	{
		if(!i->first->is_visible())
			continue;

		const Icon &icon = *i->second;

		GL::Vector3 p = renderer.get_camera()->project(i->first->get_node());
		p.x = int(p.x*0.5*fb_width-icon.width*size/2);
		p.y = int(p.y*0.5*fb_height);

		GL::Renderer::Push push2(renderer);
		renderer.matrix_stack() = GL::Matrix::translation(p);
		renderer.matrix_stack() *= GL::Matrix::scaling(size);

		icon.background.draw(renderer);

		glColor3f(0.0, 1.0, 0.0);
		for(vector<const GL::Mesh *>::const_iterator j=icon.graphics.begin(); j!=icon.graphics.end(); ++j)
		{
			(*j)->draw(renderer);
			renderer.matrix_stack() *= GL::Matrix::translation(spacing, 0, 0);
		}

		renderer.matrix_stack() *= GL::Matrix::translation(0, baseline, 0);
		renderer.set_texture(&font.get_texture());
		icon.text.draw(renderer);
	}

	glColor3f(1.0, 1.0, 1.0);
}

Overlay3D::Icon &Overlay3D::get_icon(const Object3D &track)
{
	Icon *&icon = icons[&track];
	if(!icon)
		icon = new Icon;

	return *icon;
}

const GL::Mesh *Overlay3D::get_graphic(const string &name)
{
	GL::Mesh *&grf = graphics[name];
	if(!grf)
	{
		grf = new GL::Mesh;
		try
		{
			DataFile::load(*grf, (FS::Path("icons")/(name+".mesh")).str());
		}
		catch(const exception &e)
		{
			IO::print("Error loading overlay graphic '%s': %s\n", name, e.what());
			delete grf;
			grf = 0;
		}
	}

	return grf;
}

void Overlay3D::update_icon(Icon &icon)
{
	icon.background.clear();

	icon.width = max(icon.graphics.size()*1.1+font.get_string_width(icon.label), 1.0);

	{
		GL::MeshBuilder bld(icon.background);
		bld.color(0.2f, 0.2f, 0.2f, 0.7f);

		bld.begin(GL::TRIANGLE_FAN);
		bld.vertex(0.4, 0.5);
		bld.vertex(icon.width-0.4, 0.5);
		bld.vertex(icon.width-0.4, 1.2);
		for(int i=4; i<=12; ++i)
			bld.vertex(0.4+cos(i*M_PI/8)*0.7, 0.5+sin(i*M_PI/8)*0.7);
		bld.end();

		bld.begin(GL::TRIANGLE_FAN);
		bld.vertex(icon.width-0.4, 0.5);
		bld.vertex(0.4, 0.5);
		bld.vertex(0.4, -0.2);
		for(int i=-4; i<=4; ++i)
			bld.vertex(icon.width-0.4+cos(i*M_PI/8)*0.7, 0.5+sin(i*M_PI/8)*0.7);
		bld.end();
	}

	icon.text.clear();

	{
		GL::MeshBuilder bld(icon.text);
		bld.color(0.0f, 1.0f, 0.0f);
		font.build_string(icon.label, bld);
	}
}


Overlay3D::Icon::Icon():
	width(1.0),
	background((GL::COLOR4_UBYTE, GL::VERTEX2)),
	text((GL::COLOR4_UBYTE, GL::TEXCOORD2, GL::VERTEX2))
{ }

} // namespace R2C2
