#include <cmath>
#include <msp/gl/meshbuilder.h>
#include <msp/gl/renderer.h>
#include <msp/input/keys.h>
#include <msp/strings/format.h>
#include "designer.h"
#include "3d/layout.h"
#include "measure.h"

using namespace std;
using namespace R2C2;
using namespace Msp;

Measure::Measure(Designer &d, Input::Keyboard &k, Input::Mouse &m):
	Tool(d, k, m),
	start_pinned(false),
	mesh((GL::COLOR4_UBYTE, GL::VERTEX3))
{
	update_mesh();

	designer.get_layout_3d().get_scene().add(*this);
}

Measure::~Measure()
{
	designer.get_layout_3d().get_scene().remove(*this);
}

void Measure::back()
{
	if(start_pinned)
	{
		start_pinned = false;
		update_mesh();
	}
	else
		set_done(true);
}

void Measure::key_press(unsigned key)
{
	if(key==Input::KEY_ESC)
		back();
	else
		Tool::key_press(key);
}

void Measure::button_press(unsigned btn)
{
	if(btn==1)
	{
		start.position = ground_pointer;
		start.rotation = Angle::zero();
		snap_to_tracks(start);

		start_pinned = true;
	}
	else if(btn==3)
		back();
}

void Measure::pointer_motion()
{
	Snap sn = start;
	sn.position = ground_pointer;
	snap_to_tracks(sn);
	ground_pointer = sn.position;

	if(start_pinned)
	{
		Vector delta = rotated_vector(ground_pointer-start.position, -start.rotation);

		par_dist = delta.x;
		perp_dist = delta.y;

		adiff = wrap_balanced(sn.rotation-start.rotation+Angle::half_turn());

		update_mesh();

		string info = format("Par %.1fmm - Perp %.1fmm - Total %.1fmm - Angle %.1f°", par_dist*1000, perp_dist*1000, delta.norm()*1000, adiff.degrees());
		signal_status.emit(info);
	}
}

void Measure::render(GL::Renderer &renderer, const GL::Tag &) const
{
	GL::Renderer::Push push(renderer);
	const Vector &pos = (start_pinned ? start.position : ground_pointer);
	renderer.matrix_stack() *= GL::Matrix::translation(pos);

	mesh.draw(renderer);
}

void Measure::update_mesh()
{
	mesh.clear();
	GL::MeshBuilder bld(mesh);
	bld.color(1.0f, 1.0f, 1.0f, 1.0f);
	bld.begin(GL::QUAD_STRIP);
	for(unsigned i=0; i<=16; ++i)
	{
		float x = cos(i*M_PI/8)*0.005;
		float y = sin(i*M_PI/8)*0.005;
		bld.vertex(x, y, 0);
		bld.vertex(x, y, 0.01);
	}
	bld.end();

	if(start_pinned)
	{
		float c = cos(start.rotation);
		float s = sin(start.rotation);
		bld.begin(GL::QUAD_STRIP);
		bld.vertex(0, 0, 0);
		bld.vertex(0, 0, 0.01);
		bld.vertex(c*par_dist, s*par_dist, 0);
		bld.vertex(c*par_dist, s*par_dist, 0.01);
		bld.vertex(ground_pointer.x-start.position.x, ground_pointer.y-start.position.y, 0);
		bld.vertex(ground_pointer.x-start.position.x, ground_pointer.y-start.position.y, 0.01);
		bld.vertex(0, 0, 0);
		bld.vertex(0, 0, 0.01);
		bld.end();
	}
}

void Measure::snap_to_tracks(Snap &sn)
{
	const set<Track *> &ltracks = designer.get_layout().get_all<Track>();
	for(set<Track *>::const_iterator i=ltracks.begin(); i!=ltracks.end(); ++i)
		if((*i)->snap(sn, 0.01, SNAP_NODE))
			return;
}
